local AddonName, GGL = ...
local Templates = GGL.Templates
local Sender = GGL.Sender
local BNPlayer = Templates.BNPlayer
local BNWoWPlayer = Templates.BNWoWPlayer

local Sync, SyncHelper = Templates:Register("Sync")

-- lua
local paris, ipairs = pairs, ipairs
local type = type
local match, gmatch, sub = string.match, string.gmatch, string.sub
local wipe, concat = table.wipe, table.concat

-- wow
local C_TimerAfter = C_Timer.After

-- ##################
-- local
-- ##################
local SyncRegister = {}
local PlayerRegister = {}

local INVITE_STATES = GGL.INVITE_KEY_LIST

-- no same interval here
local LOOT_SYNC_INTERVAL = 2
local SPECIAL_DATA_SYNC_INTERVAL = 5
local MISSING_LOOT_COOLDOWN = 10

local MAX_LOOT_ITEMS_SEND_WOW = 10
local MAX_LOOT_ITEMS_SEND_BN = 100

local ID_MATCH_FORMAT = {
    sync =      "^(.*):(.*)",   -- battleTag:UnitGUID
    char =      "^(.*)%-(.*)",  -- name:server
    guid =      "^Player.*",    -- UnitGUID
}

local function GetPlayer(player, allowDummy)
    if not player or player == "" then return end

    -- game account ID
    -- gameAccountID is fixed for a battleNet+WoWAccount for a login session
    if type(player) == "number" then
        local newPlayer = BNWoWPlayer:GetByGameAccountID(player)
        if newPlayer then
            return newPlayer
        end
        return false, "GameAccountID"
    end

    if PlayerRegister[player] then
        return PlayerRegister[player]
    end

    if type(player) == "table" then
        if player.GetType then
            PlayerRegister[player] = player
            return player
        else
            return false, "UnknownTable"
        end
    end

    -- sync name string
    local a, b = match(player, ID_MATCH_FORMAT.sync)
    if a and b then
        local newPlayer = BNWoWPlayer:GetByGUID(b)
        if newPlayer then
            PlayerRegister[player] = newPlayer
            return newPlayer
        elseif allowDummy then
            return BNWoWPlayer:GetDummyPlayer(player)
        end
        return false, "UnitGUID"
    end

    -- wow player name
    a, b = match(player, ID_MATCH_FORMAT.char)
    if a and b then
        local newPlayer = Templates:Get("WoWPlayer", player)
        if newPlayer then
            PlayerRegister[player] = newPlayer
            return newPlayer
        end
        return false, "WoWPlayer"
    end
end

-- set timers only one time
local TimerRegister = {}
local function Set_C_TimerAfter(after, method)
    if not TimerRegister[method] then
        TimerRegister[method] = true
        C_TimerAfter(after, method)
    end
end

local function GetSenderData(prefix, sender, prefixSize)
    local session = SyncRegister[sub(prefix, 1, -(1 + (prefixSize or 1)))]
    if not session then return end
    -- check for player
    sender = GetPlayer(sender)
    if not sender then return end
    local syncID = session:GetPlayerSyncID(sender)
    if not syncID then return end

    return session, sender, syncID
end

-- <session>.senderHandler      CB
local function OnSenderMessageReceive(self, prefix, sender, message, isBNMessage)
    local session = SyncRegister[sub(prefix, 1, -2)]
    if not session then return end
    -- only players that accepted can send loot
    -- if session:GetPlayerInviteState(sender) ~= INVITE_STATES.ACCEPTED then return end
    sender = GetPlayer(sender)
    if not sender then return GGL:AddIntoErrorLog("OnSenderMessageReceive", {session.sessionID, session:GetPlayerSyncID(sender)}) end

    local pass, lootData = GGL:Deserialize(message)
    if pass then
        for i = 1, #lootData do
            session:AddLootString(sender, lootData[i][2], lootData[i][1])
        end
    end
end

-- <session>.senderGetMissingLoot      CB
local function OnGetMissingLootMessageReceive(self, prefix, sender, message, isBNMessage)
    -- check for session
    local session = SyncRegister[sub(prefix, 1, -2)]
    if not session then return end
    -- check for player
    sender = GetPlayer(sender)
    if not sender or not session.lootLog then return end

    -- check if its just a sync wth no specific ID's
    local count = match(message, "^Count\002(%d+)")
    if count then
        count = tonumber(count)
        -- compare count with num lootEntrys
        if count == #session.lootLog then
            -- we are fine, do nothing
        else
            -- num items if off send back count
            session.senderSendMissingLoot:SendMessage(sender, "Count\002"..#session.lootLog)
        end
    else
        -- we get a list of missing ID's
        -- sendID:id1:id2:id3:...   -- : = \002
        local lootDB = session.lootLog
        local lootTable = {}
        for id in gmatch(message, "%d+") do
            id = tonumber(id)
            if lootTable.sendID then
                lootTable[#lootTable + 1] = {id,lootDB[id]}
            else
                lootTable.sendID = id
            end
        end
        session:SendLootMessageToHost(lootTable, true)
    end
end

-- <session>.senderSendMissingLoot      CB
local function OnReceiveMissingLootMessageReceive(self, prefix, _sender, message, isBNMessage)
    local session, sender, syncID = GetSenderData(prefix, _sender)
    if not syncID or not session.db.lootLog[syncID] then return GGL:AddIntoErrorLog("OnReceiveMissingLootMessageReceive", {session.sessionID, syncID}) end

    -- we get back a count when its different from ours just run a new check
    local count = match(message, "^Count\002(%d+)")

    if count then
        if session.db.lootLog[syncID] then
            -- we get back a count as its different from ours just run a new check
            session.db.lootLog[syncID].count = tonumber(count)
            session:RequestMissingLoot(sender)
        end
    else
        local pass, lootData = GGL:Deserialize(message)
        if pass then
            for index, loot in ipairs(lootData) do
                session:AddLootString(sender, loot[2], loot[1], lootData.sendID)
            end
        end
    end
end

-- <session>.senderSendSpecialData      CB
local function OnReceiveSpecialDataMessageReceive(self, prefix, _sender, message, isBNMessage)
    local session, sender, syncID = GetSenderData(prefix, _sender, 2)
    if not syncID then return end

    local pass, specialData = GGL:Deserialize(message)
    if pass then
        -- kill count
        session.db.playerKillCount[syncID] = specialData.k
    end
end

-- SessionEvent "SessionTemplate_AddLogItem"         CB
local function OnNewLogEntry(sessionSelf, playerTemplate, logString, logCounter)
    if not sessionSelf.hostPlayer then return end
    sessionSelf.lootSendList.count = ((sessionSelf.lootSendList.count or 0) > logCounter) and (sessionSelf.lootSendList.count or 0) or logCounter
    sessionSelf.lootSendList[#sessionSelf.lootSendList+1] =  {logCounter, logString}
end

-- Checks for missiong loot from a given player
-- @param       self        <sessionTemplate>
-- @param       player      <playerTemplate>
local function CheckMissingLoot(self, player)
    if not self.db.lootLog[self:GetPlayerSyncID(player)] then return end
    local log = self.db.lootLog[self:GetPlayerSyncID(player)]
    local count = log.count
    if count < #log then
        count = #log
    end

    local time = GGL.GetTime()
    local missingLoot = {time}
    for i = 1, count do
        if not log[i] or (type(log[i]) == "number" and (time - log[i]) > MISSING_LOOT_COOLDOWN) then
            log[i] = time
            missingLoot[#missingLoot+1] = i
        end
    end

    if #missingLoot > 1 then
        self.senderGetMissingLoot:SendMessage(player, concat(missingLoot, "\002"))
    end
end

-- ## Invite Handler ##

--- Fires when the invite state changes
local function CB_it_PlayerInviteStateUpdate(inviteTemplate, playerTemplate, inviteState)
    local session = inviteTemplate.session
    if not session then return end
    if inviteState == INVITE_STATES.ACCEPTED then
        session:FireCallbackEvent("SessionTemplate_Sync_PlayerAccepted", session, playerTemplate)
        session:SendStart(playerTemplate)
        if session.db.activityLog then
            session.db.activityLog[#session.db.activityLog+1] = { playerTemplate.syncID, GGL.GetTime() > session.db.endTime and session.db.endTime or GGL.GetTime(), 1 }
        end
    elseif inviteState == INVITE_STATES.DECLINED then
        session:FireCallbackEvent("SessionTemplate_Sync_PlayerDeclined", session, playerTemplate)
        if session.db.activityLog then
            session.db.activityLog[#session.db.activityLog+1] = { playerTemplate.syncID, GGL.GetTime() > session.db.endTime and session.db.endTime or GGL.GetTime(), 2 }
        end
    elseif inviteState == INVITE_STATES.INVITED then
        session:FireCallbackEvent("SessionTemplate_Sync_PlayerInvited", session, playerTemplate)
    end
end

--- Fires the "SessionTemplate_Sync_NewSyncLoot" event
local function NewSyncLoot(self, player, lootString)
    if not player or type(player) ~= "table" then return end
    local lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag = self:GetDataFromLogString(lootString)

    lootTime = lootTime + player:GetTimeOffset()
    -- check if item is cached
    if itemID == 0 then
        self:FireCallbackEvent("SessionTemplate_Sync_NewSyncLoot", self, player, lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag)
    else
        GGL:IsItemCached(itemID, self.FireCallbackEvent, self, "SessionTemplate_Sync_NewSyncLoot", self, player, lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag)
    end
end

-- ##################
-- static
-- ##################
local Static = {}
Templates.Sync = Static

Static.GetPlayerFromSync = GetPlayer

-- ##################
-- template

-- EVENTS
--      "OnRemovePlayer"        (sessionSelf, playerTemplate)        -- player removed from session
--      "OnAddPlayer"           (sessionSelf, playerTemplate)        -- player added into PlayerTemplateList
--      "SessionTemplate_Sync_NewSyncLoot"           (sessionSelf, player, lootTime, zoneIniID, difficultyID, itemID, ammount, sourceID, mainSourceFlag, subSourceFlag)
--      "SessionTemplate_Sync_PlayerAccepted"        (sessionSelf, playerTemplate)       -- player accepted a invite
--      "SessionTemplate_Sync_PlayerDeclined"        (sessionSelf, playerTemplate)       -- player declined a invite
--      "SessionTemplate_Sync_PlayerInvited"         (sessionSelf, playerTemplate)       -- player is invited
--      "SessionTemplate_Sync_PlayerMergeIDUpdate"   (sessionSelf)                       -- a player in lootLog gets a mergeID update
-- ##################
Sync.isSyncRegistered = true
function Sync:Constructor()
    local db = self.db
    if not db then return end

    -- add into sync list
    SyncRegister[self.db.sessionID] = self

    -- contains loot that get send to host
    self.lootSendList = {}

    -- get existing playerlist if aviable
    db.playerList = db.playerList or {}

    -- mergeID saver
    db.mergeIDList = db.mergeIDList or {}

    -- sync needs a inviteTemplate
    self.inviteTemplate = Templates:New("Invite", self, false)
    self.inviteTemplate:SetPlayerDBList(db.playerList, true)
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", CB_it_PlayerInviteStateUpdate)

    -- create function for accountID updates
    if not self.__playerAcountIDUpdate then
        self.__playerAcountIDUpdate = function(blPlayerStatic, updatetSyncIDs)
            -- update mergeID saver
            for syncID in pairs(db.mergeIDList) do
                local newPlayer, err = GetPlayer(syncID)
                if newPlayer then
                    self:GetPlayerMergeID(newPlayer)
                end
            end
            self:UpdateSyncInstanceResetData()
            -- fire callback
            self:FireCallbackEvent("SessionTemplate_Sync_PlayerMergeIDUpdate", self)
        end
    end

    -- prepear loot sender
    self._LootSyncTimerCallFunction = function()
        if not self.hostPlayer or db.isClosed or not self.isRunning then
            TimerRegister[self._LootSyncTimerCallFunction] = false
            return
        end
        self:SendLootMessageToHost()
        C_TimerAfter(LOOT_SYNC_INTERVAL, self._LootSyncTimerCallFunction)
    end

    -- prepear special data sender
    self._SpecialDataTimerCallFunction = function()
        if not self.hostPlayer or db.isClosed or not self.isRunning then
            TimerRegister[self._SpecialDataTimerCallFunction] = false
            return
        end
        self:SendOwnSpecialDataToHost()
        C_TimerAfter(SPECIAL_DATA_SYNC_INTERVAL, self._SpecialDataTimerCallFunction)
    end
end

function Sync:UpdateSyncInstanceResetData()
    if not self.db.instanceResetData then return end
    local iniResetData = self.db.instanceResetData
    for syncID in pairs(self.db.lootLog) do
        local playerTemplate = GetPlayer(syncID, true)
        if playerTemplate and playerTemplate.mergeID and iniResetData[syncID] then
            local tmp = iniResetData[syncID]
            iniResetData[syncID] = nil
            self:UpdateResetData(playerTemplate, tmp)
        end
    end
end

--- Set a seralized settings string
-- @param       setting     <seralized:string>
function Sync:SetSettingsString(settings)
    if not settings then return end
    local pass, settingsData = GGL:Deserialize(settings)
    if pass then
        local db = self.db
        db.setSave = db.setSave or {}
        -- load settings
        for settingKey, settingString in pairs(settingsData) do
            if db.setSave[settingKey] then
                db.setSave[settingKey] = settingString
            end
        end
        self:ReloadSet()
    end
end

--- Set the session timer from the host
-- @param       fullDuration        <number:serverTime> Full furation of the session
-- @param       elapsed             <number> Elapsed time in seconds
function Sync:SetTimeFromSync(fullDuration, elapsed)
    local curTime = GGL.GetTime()
    self.durationTime = fullDuration
    self.startTime = curTime - elapsed
    self.endTime = self.startTime + fullDuration
end

--- Returns the merge ID for a playerTemplate
-- @param       playerTemplate      <playerTemplate>
-- @return  <mergeID or nil or "self"(for player)>
function Sync:GetPlayerMergeID(playerTemplate)
    if not playerTemplate or playerTemplate.mergeID == 0 then
        return nil
    else
        self.db.mergeIDList[playerTemplate.syncID] = playerTemplate.mergeID
        return playerTemplate.mergeID
    end
end

--- Returns the active player for a mergeID
-- @param       mergeID      <playerTemplate.mergeID>
-- @return  playerTemplate
function Sync:GetActivePlayerForMergeID(mergeID)
    for playerTemplate, inviteState in self:IterAcceptedPlayers() do
        if playerTemplate.mergeID == mergeID and playerTemplate.isOnline then
            return playerTemplate
        end
    end
end

--- Get the sessionSyncID for a player
-- this overrites the basic one..
-- @param       playerTemplate      <playerTemplate>
-- @param       addIntoLog          <bool> add into the lootLog table
function Sync:GetPlayerSyncID(playerTemplate, addIntoLog)
    if not self.db.lootLog then return end

    if playerTemplate == self.selfPlayer then
        return self:CallBaseFunction("Session", "GetPlayerSyncID")
    end

    -- player is not in merge and loot list
    -- TODO: #9 Add more merge settings here
    if addIntoLog and not self.db.lootLog[playerTemplate.syncID] then
        self.db.lootLog[playerTemplate.syncID] = {}
    end

    return playerTemplate.syncID
end

--- Adds a loot string into the loot log
-- @param       player          <playerTemplate>
-- @param       lootString      <string> full loot string
-- @param       count           <number> lootLog number
-- @param       sendID          <number> the given ID of that sync
function Sync:AddLootString(playerTemplate, lootString, count, sendID)
    if not playerTemplate or not lootString or not count then return GGL:AddIntoErrorLog("Sync:AddLootString / missing", {self.sessionID, playerTemplate and self:GetPlayerSyncID(playerTemplate) or "playerTemplate", count, sendID, lootString}) end

    local log = self.db.lootLog[self:GetPlayerSyncID(playerTemplate, true)]
    if sendID and log[count] and log[count] ~= sendID then return GGL:AddIntoErrorLog("Sync:AddLootString", {self.sessionID, self:GetPlayerSyncID(playerTemplate), count, sendID, lootString}) end -- no double/wrong entrys...

    -- set lootEntry and logSize
    log[count] = lootString
    if not log.count or count > log.count then
        log.count = count
    end

    -- add into session
    NewSyncLoot(self, playerTemplate, lootString)
end

--- Request missing loot from a player
-- @param       player      <playerTemplate> if player is true all players get checked
function Sync:RequestMissingLoot(player)
    if not self:PlayerIsHost() then return end
    if player == true then
        -- player true means just send count
        for playerTemplate, inviteState in self:IterAcceptedPlayers() do
            if self.db.lootLog[self:GetPlayerSyncID(playerTemplate)] then
                self.senderGetMissingLoot:SendMessage(playerTemplate, "Count\002"..#self.db.lootLog[self:GetPlayerSyncID(playerTemplate)])
            end
        end
    else
        if player then
            CheckMissingLoot(self, player)
        else
            for playerTemplate in self:IterAcceptedPlayers() do
                CheckMissingLoot(self, playerTemplate)
            end
        end
    end
end

--- Sets a player as host
-- @param       player                  <playerTemplate>
-- @param       ignoreInviteTemplate    <bool> if true the host not gets set for the inviteTemplate
function Sync:SetHostPlayer(player, ignoreInviteTemplate)
    if player and self.hostPlayer == player then return end
    if not ignoreInviteTemplate then
        return self.inviteTemplate:SetHost(player)
    end

    -- set host
    self.hostPlayer = player

    -- set player as host if none found
    if not self.hostPlayer then
        self.hostPlayer = self.selfPlayer
    end

    -- db
    self.db.hostPlayer = self.hostPlayer and self:GetPlayerSyncID(self.hostPlayer) or nil
    self.db.playerIsHost = self.hostPlayer == self.selfPlayer
end

--- Get the host playerTemplate
-- @return hostPlayerTemplate
function Sync:GetHostPlayer()
    return self.hostPlayer
end

-- Check if the given player is a valid host for that session
-- @param       playerTemplate      <playerTemplate>
-- @return isValidHost
function Sync:IsValidHost(playerTemplate)
    if not playerTemplate then return end

    if not self.db.hostPlayer then
        -- no host found so its valid
        return true
    elseif self.hostPlayer and playerTemplate.bnPlayer.battleTag == self.hostPlayer.bnPlayer.battleTag then
        -- same battleTag so host
        return true
    elseif self.db.playerIsHost and playerTemplate == self.selfPlayer then
        -- player is host
        return true
    elseif self.db.hostPlayer == self:GetPlayerSyncID(playerTemplate) then
        -- same player
        return true
    elseif playerTemplate.bnPlayer and (match(self.db.hostPlayer, ID_MATCH_FORMAT.sync) == playerTemplate.bnPlayer.battleTag) then
        -- same battleTag
        return true
    end
    return false
end

--- Check if the player self is a valid host
-- @return isValidHost
function Sync:PlayerIsValidHost()
    return self:IsValidHost(self.selfPlayer)
end

--- Set if settings should get synced when change
-- @param       state       <bool>
function Sync:SyncSettingsWithPlayers(state)
    self.db.syncSettings = state and true or nil
end

--- Sends collected loot to host
-- @param       lootTable       <table>
-- @param       isMissingLoot   <bool> Set if the loot was missing loot from host
function Sync:SendLootMessageToHost(lootTable, isMissingLoot)
    if not self.hostPlayer then return end
    lootTable = lootTable or self.lootSendList
    if #lootTable <= 0 then return end
    local sendCap
    local sendHandler = isMissingLoot and self.senderSendMissingLoot or self.senderHandler
    if not sendHandler or not sendHandler.SendMessage then return end

    -- get split size for tables or it gets spammy
    if self.hostPlayer:GetType() == "BNWoWPlayer" and #lootTable > MAX_LOOT_ITEMS_SEND_BN then
        sendCap = MAX_LOOT_ITEMS_SEND_BN
    elseif self.hostPlayer:GetType() == "WoWPlayer" and #lootTable > MAX_LOOT_ITEMS_SEND_WOW then
        sendCap = MAX_LOOT_ITEMS_SEND_WOW
    end
    -- split messages if we are over the cap
    if sendCap then
        local send = {
            sendID = lootTable.sendID
        }
        for index, logItem in ipairs(lootTable) do
            send[#send+1] = logItem
            if #send == sendCap then
                sendHandler:SendMessage(self.hostPlayer, GGL:Serialize(send))
                send = {
                    sendID = lootTable.sendID
                }
            end
        end
        if #send > 0 then
            sendHandler:SendMessage(self.hostPlayer, GGL:Serialize(send))
        end
    else
        sendHandler:SendMessage(self.hostPlayer, GGL:Serialize(lootTable))
    end
    -- reset table
    wipe(lootTable)
end

--- Sends special data to host
-- contains:
--  k = killCount
function Sync:SendOwnSpecialDataToHost()
    if not self.hostPlayer then return end
    if not self.senderSendSpecialData then return end
    self.senderSendSpecialData:SendMessage(self.hostPlayer, GGL:Serialize({
        k = self.db.playerKillCount[self.selfDbKey] or 0,    -- kill count
    }))
end

--- Get inviteTemplate
-- @return inviteTemplate
function Sync:GetInviteTemplate()
    return self.inviteTemplate
end

--- Sets a new description for the session
-- @param       description         <string>
function Sync:SetDescription(description)
    self.description = description
end

--- Opens battleNet invites from inviteTemplate
-- @param       state       <bool>
function Sync:OpenBNInvites(state)
    self.inviteTemplate:OpenBattleNetInvites(state)
end

--- Get the inviteState for a player
-- @param       player      <playerTemplate/string> check local function GetPlayer
function Sync:GetPlayerInviteState(player)
    local newPlayer, err = GetPlayer(player)
    if not newPlayer or err then
        -- Player not found
        -- print(err)
    else
        if self.hostPlayer == newPlayer then
            return INVITE_STATES.ACCEPTED
        else
            return self.db.playerList[self:GetPlayerSyncID(newPlayer)]
        end
    end
end

--- Get the inviteStateList from the db
-- @return inviteStateList<table>
function Sync:GetInviteStateList()
    return self.db.playerList
end

--- Check if player is host
-- @return playerIsHost
function Sync:PlayerIsHost()
    return self.db.playerIsHost
end

-- for player, inviteState in Sync:IterInvitedPlayers() do end
function Sync:IterInvitedPlayers()
    local playerList = {}
    for name, state in pairs(self:GetInviteStateList()) do
        local player = GetPlayer(name)
        if player and type(player) == "table" and not player.isPlayer then
            playerList[GetPlayer(name)] = state
        end
    end
    return pairs(playerList)
end

-- for playerTemplate, inviteState in Sync:IterAcceptedPlayers() do end
function Sync:IterAcceptedPlayers()
    local playerList = {}
    for name, state in pairs(self:GetInviteStateList()) do
        if state == INVITE_STATES.ACCEPTED then
            local player = GetPlayer(name)
            if player and type(player) == "table" and not player.isPlayer then
                playerList[GetPlayer(name)] = state
            end
        end
    end
    return pairs(playerList)
end

--- Sends the session start to a player
-- @param       playerTemplate      <playerTemplate>
function Sync:SendStart(playerTemplate)
    if not self.inviteTemplate then return end

    if playerTemplate then
        self.inviteTemplate:SendStartSession(playerTemplate)
    else
        self.inviteTemplate:StartSession()
        if not self.db.activityLog then
            self.db.activityLog = {}
            for playerTemplate, inviteState in self:IterAcceptedPlayers() do
                self.db.activityLog[#self.db.activityLog+1] = { playerTemplate.syncID, self.db.startTime, 1 }
            end
        end
    end
end

--- Reloads the settingsSets for a session
function Sync:ReloadSet()
    self:CallBaseFunction("Session", "ReloadSet")
end

--- Rebuilds the full loot list from all players and fires the events
function Sync:RebuildFromLog()
    GGL.LootAlert:Pause(true)
    for playerSyncID, lootData in pairs(self.db.lootLog) do
        -- ignore empty tables
        if lootData.count then
            local player = GetPlayer(playerSyncID, true)
            -- ignore player (gets loaded with main session)
            if not player.isPlayerBattleTag then
                local rebuild
                for i = 1, lootData.count do
                    if lootData[i] and type(lootData[i]) == "string" then
                        NewSyncLoot(self, player, lootData[i])
                    else
                        rebuild = true
                    end
                end
                if rebuild then
                    self:RequestMissingLoot(player)
                end
            end
        end
    end
    GGL.LootAlert:Pause(false)
end

--- Override as synced session should not change startTime
function Sync:SetTimerEndless()
    if self.db.startTime then
        self.db.endlessTime = true
        self.db.timeElapsed = GGL.GetTime() - self.db.startTime
    else
        self.db.endlessTime = true
        self.db.timeElapsed = self.db.timeElapsed or 0
        self.db.startTime = GGL.GetTime() - self.db.timeElapsed
    end
end

--- Starts the session
function Sync:Start()
    if not self.db.isRunning then return end
    local db = self.db

    -- add into sync list
    SyncRegister[self.db.sessionID] = self

    -- register sender
    if not self.senderHandler or not self.senderHandler.Unregister then
        self.senderHandler = GGL.Sender:RegisterReceiver(self.db.sessionID.."l", OnSenderMessageReceive)
        self.senderGetMissingLoot = GGL.Sender:RegisterReceiver(self.db.sessionID.."g", OnGetMissingLootMessageReceive)
        self.senderSendMissingLoot = GGL.Sender:RegisterReceiver(self.db.sessionID.."s", OnReceiveMissingLootMessageReceive)
        self.senderSendSpecialData = GGL.Sender:RegisterReceiver(self.db.sessionID.."sd", OnReceiveSpecialDataMessageReceive)
    end

    -- register log events
    self:RegisterCallbackEvent("SessionTemplate_AddLogItem", OnNewLogEntry)
    self:RegisterCallbackEvent("SessionTemplate_AddLogMoney", OnNewLogEntry)

    -- register inviteTemplate events
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", CB_it_PlayerInviteStateUpdate)

    --self:OpenBNInvites(true)
    self:SendStart()
    self:RequestMissingLoot(true)
    -- rebuild list from db
    if not self.rebuildSync then
        self:RebuildFromLog()
        self.rebuildSync = true
    end

    -- set accountID updates
    Templates.BNPlayer:RegisterCallbackEvent("PlayerAcountIDUpdate", self.__playerAcountIDUpdate)

    -- set loot sync
    if not self.db.playerIsHost then
        Set_C_TimerAfter(LOOT_SYNC_INTERVAL, self._LootSyncTimerCallFunction)
        Set_C_TimerAfter(SPECIAL_DATA_SYNC_INTERVAL, self._SpecialDataTimerCallFunction)
    end

    -- update ini resets
    self:UpdateSyncInstanceResetData()
end

local function UnregisterAll(self)
    self:UnregisterAllCallbackEvents()

    -- unregister sender
    if self.senderHandler and self.senderHandler.Unregister then
        self.senderHandler:Unregister()
        self.senderGetMissingLoot:Unregister()
        self.senderSendMissingLoot:Unregister()
        self.senderSendSpecialData:Unregister()
    end

    Templates.BNPlayer:UnregisterCallbackEvent("PlayerAcountIDUpdate", self.__playerAcountIDUpdate)

    SyncRegister[self.db.sessionID] = nil
end

--- Unregister all relevant things from that session
function Sync:Unregister()
    UnregisterAll(self)
end

--- Stops the session
function Sync:Stop()
    if self.inviteTemplate then
        self.inviteTemplate:Stop()
    end
    UnregisterAll(self)

    if not self.inviteTemplate then return end
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", CB_it_PlayerInviteStateUpdate)
    self.inviteTemplate:SetOwnInviteState(INVITE_STATES.DECLINED)
    SyncRegister[self.db.sessionID] = self
    self.rebuildSync = nil
end

--- Close the session
function Sync:Close()
    if self.inviteTemplate then
        self.inviteTemplate:Close()

        self.inviteTemplate = nil
    end
    UnregisterAll(self)
end